@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c), 2016-2021, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -189,11 +189,11 @@ class TestAbstractGenerator(unittest.TestCase):
 
         self.assertIs(generator.schema, self.schema)
         self.assertIsInstance(generator._env, jinja2.Environment)
-        self.assertEqual(repr(generator), "{}(namespace={!r})".format(class_name, namespace))
+        self.assertEqual(repr(generator), f"{class_name}(namespace={namespace!r})")
 
         generator = self.generator_class(self.col_xsd_file)
         self.assertIsInstance(generator.schema, XMLSchema11)
-        self.assertEqual(repr(generator), "{}(schema='collection.xsd')".format(class_name))
+        self.assertEqual(repr(generator), f"{class_name}(schema='collection.xsd')")
 
     def test_searchpath_argument(self):
         class DemoGenerator2(AbstractGenerator):
@@ -239,7 +239,7 @@ class TestAbstractGenerator(unittest.TestCase):
         template_dir = Path(__file__).parent.joinpath('templates')
         language = self.generator_class.formal_language.lower()
 
-        templates = set(x.name for x in template_dir.glob('{}/*'.format(language)))
+        templates = {x.name for x in template_dir.glob(f'{language}/*')}
         templates.update(x.name for x in template_dir.glob('filters/*'))
         self.assertSetEqual(set(self.generator.list_templates()), templates)
 
@@ -535,8 +535,8 @@ class TestAbstractGenerator(unittest.TestCase):
 
     def test_is_derivation(self):
         self.assertFalse(self.generator.extension(self.schema.types['type1']))
-        self.assertFalse(self.generator.extension(self.schema.types['type1'], 'tns:type1'))
-        self.assertFalse(self.generator.restriction(self.schema.types['type1'], 'tns:type1'))
+        self.assertTrue(self.generator.extension(self.schema.types['type1'], 'tns:type1'))
+        self.assertTrue(self.generator.restriction(self.schema.types['type1'], 'tns:type1'))
         self.assertTrue(self.generator.derivation(self.schema.types['type1'], 'type1'))
         self.assertFalse(self.generator.restriction(self.schema.types['type6'], 'xs:decimal'))
         self.assertFalse(self.generator.restriction(self.schema.types['type6'], None))
@@ -634,9 +634,5 @@ class TestPythonGenerator11(TestPythonGenerator):
 
 
 if __name__ == '__main__':
-    import platform
-    header_template = "Test xmlschema code generators with Python {} on {}"
-    header = header_template.format(platform.python_version(), platform.platform())
-    print('{0}\n{1}\n{0}'.format("*" * len(header), header))
-
-    unittest.main()
+    from xmlschema.testing import run_xmlschema_tests
+    run_xmlschema_tests('code generators')
