@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -8,7 +8,8 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
-import unittest
+import warnings
+from pathlib import Path
 from textwrap import dedent
 from xml.etree.ElementTree import Element
 
@@ -19,6 +20,8 @@ from xmlschema.testing import XsdValidatorTestCase
 
 class TestXsdComplexType(XsdValidatorTestCase):
 
+    cases_dir = Path(__file__).parent.joinpath('../test_cases')
+
     def check_complex_restriction(self, base, restriction, expected=None, **kwargs):
         content = 'complex' if self.content_pattern.search(base) else 'simple'
         source = """
@@ -588,6 +591,43 @@ class TestXsdComplexType(XsdValidatorTestCase):
                 </xs:complexType>
             </xs:schema>"""))
 
+    def test_mixed_content_extension__issue_414(self):
+        # Not a bug, the user refers to an old version (v1.10), but
+        # there is a detailed analysis on that:
+        #   https://stackoverflow.com/a/78942158/1838607
+        xsd_file = self.casepath('issues/issue_414/issue_414.xsd')
+        xml_file = self.casepath('issues/issue_414/issue_414.xml')
+
+        schema = self.schema_class(xsd_file)
+        self.assertTrue(schema.types['mixedElement'].mixed)
+        self.assertTrue(schema.elements['root'].type.mixed)
+        self.assertTrue(schema.is_valid(xml_file))
+
+        xsd_file = self.casepath('issues/issue_414/issue_414b.xsd')
+        schema = self.schema_class(xsd_file)
+        self.assertTrue(schema.types['mixedElement'].mixed)
+        self.assertTrue(schema.elements['root'].type.mixed)
+        self.assertTrue(schema.is_valid(xml_file))
+
+        xsd_file = self.casepath('issues/issue_414/issue_414ne.xsd')
+        schema = self.schema_class(xsd_file)
+        self.assertTrue(schema.types['mixedElement'].mixed)
+        self.assertTrue(schema.elements['root'].type.mixed)
+
+        xsd_file = self.casepath('issues/issue_414/issue_414ne-inv1.xsd')
+        with self.assertRaises(XMLSchemaParseError) as ctx:
+            self.schema_class(xsd_file)
+
+        reason = ("base has a different content type (mixed=True) "
+                  "and the extension group is not empty")
+        self.assertIn(reason, str(ctx.exception))
+
+        xsd_file = self.casepath('issues/issue_414/issue_414ne-inv2.xsd')
+        with self.assertRaises(XMLSchemaParseError) as ctx:
+            self.schema_class(xsd_file)
+
+        self.assertIn(reason, str(ctx.exception))
+
     def test_empty_content_extension(self):
         schema = self.schema_class(dedent("""\
         <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
@@ -747,6 +787,46 @@ class TestXsdComplexType(XsdValidatorTestCase):
                 </xs:complexType>
             </xs:schema>"""))
 
+    def test_illegal_restriction__issue_425(self):
+
+        schema_source = dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+
+              <xs:element name="elem1"/>
+              <xs:element name="elem2"/>
+              <xs:element name="elem3"/>
+              <xs:element name="elem4" substitutionGroup="elem3"/>
+              <xs:attribute name="a" type="xs:string"/>
+
+              <xs:complexType name="baseType">
+                <xs:sequence>
+                  <xs:element ref="elem1" minOccurs="0" maxOccurs="unbounded"/>
+                  <xs:element ref="elem2" minOccurs="0"/>
+                  <xs:element ref="elem3" maxOccurs="unbounded"/>
+                </xs:sequence>
+                <xs:attribute ref="a" use="required"/>
+              </xs:complexType>
+
+              <xs:complexType name="derivedType">
+                <xs:complexContent>
+                  <xs:restriction base="baseType">
+                    <xs:sequence>
+                      <xs:element ref="elem1" minOccurs="0" maxOccurs="unbounded"/>
+                      <xs:element ref="elem4"/>
+                    </xs:sequence>
+                    <xs:attribute ref="a" use="required"/>
+                  </xs:restriction>
+                </xs:complexContent>
+              </xs:complexType>
+            </xs:schema>""")
+
+        if self.schema_class.XSD_VERSION == '1.0':
+            with self.assertRaises(XMLSchemaParseError):
+                self.schema_class(schema_source)
+        else:
+            schema = self.schema_class(schema_source)
+            self.assertIsInstance(schema, XMLSchema11)
+
 
 class TestXsd11ComplexType(TestXsdComplexType):
 
@@ -765,8 +845,61 @@ class TestXsd11ComplexType(TestXsdComplexType):
         self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '10', 'max': '19'})))
         self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '19', 'max': '19'})))
         self.assertFalse(xsd_type.is_valid(Element('a', attrib={'min': '25', 'max': '19'})))
+
+        # doesn't set a type if it's based on a dummy element
         self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '25', 'max': '100'})))
 
+        schema = self.schema_class(dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+              <xs:complexType name="intRange">
+                <xs:attribute name="min" type="xs:int"/>
+                <xs:attribute name="max" type="xs:int"/>
+                <xs:assert test="@min le @max"/>
+              </xs:complexType>
+              <xs:element name="a" type="intRange"/>
+            </xs:schema>"""))
+
+        xsd_type = schema.types['intRange']
+        xsd_type.decode(Element('a', attrib={'min': '10', 'max': '19'}))
+        self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '10', 'max': '19'})))
+        self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '19', 'max': '19'})))
+        self.assertFalse(xsd_type.is_valid(Element('a', attrib={'min': '25', 'max': '19'})))
+
+        # doesn't set a type if it's based on a dummy element (TODO: add dummy element to maps?
+        self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '25', 'max': '100'})))
+
+        xsd_element = schema.elements['a']
+        xsd_element.decode(Element('a', attrib={'min': '10', 'max': '19'}))
+        self.assertTrue(xsd_element.is_valid(Element('a', attrib={'min': '10', 'max': '19'})))
+        self.assertTrue(xsd_element.is_valid(Element('a', attrib={'min': '19', 'max': '19'})))
+        self.assertFalse(xsd_element.is_valid(Element('a', attrib={'min': '25', 'max': '19'})))
+        self.assertTrue(xsd_element.is_valid(Element('a', attrib={'min': '25', 'max': '100'})))
+
+    def test_rooted_expression_in_assertion__issue_386(self):
+        # absolute expression in assertion
+        xsd_file = self.casepath('issues/issue_386/issue_386.xsd')
+
+        with warnings.catch_warnings(record=True) as ctx:
+            self.schema_class(xsd_file)
+            schema = self.schema_class(xsd_file)
+
+            self.assertEqual(len(ctx), 1, "Expected two assert warnings")
+            self.assertIn("absolute location path", str(ctx[0].message))
+
+        xml_file = self.casepath('issues/issue_386/issue_386-1.xml')
+        self.assertFalse(schema.is_valid(xml_file))
+        xml_file = self.casepath('issues/issue_386/issue_386-2.xml')
+        self.assertFalse(schema.is_valid(xml_file))
+
+        # relative path in assertion
+        xsd_file = self.casepath('issues/issue_386/issue_386-2.xsd')
+        schema = XMLSchema11(xsd_file)
+
+        xml_file = self.casepath('issues/issue_386/issue_386-1.xml')
+        self.assertTrue(schema.is_valid(xml_file))
+        xml_file = self.casepath('issues/issue_386/issue_386-2.xml')
+        self.assertFalse(schema.is_valid(xml_file))
+
     def test_sequence_extension(self):
         schema = self.schema_class("""
             <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
@@ -813,9 +946,5 @@ class TestXsd11ComplexType(TestXsdComplexType):
 
 
 if __name__ == '__main__':
-    import platform
-    header_template = "Test xmlschema's XSD complex types with Python {} on {}"
-    header = header_template.format(platform.python_version(), platform.platform())
-    print('{0}\n{1}\n{0}'.format("*" * len(header), header))
-
-    unittest.main()
+    from xmlschema.testing import run_xmlschema_tests
+    run_xmlschema_tests('XSD complex types')
