@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -8,16 +8,19 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
-import unittest
+import pathlib
+import xml.etree.ElementTree as ElementTree
 
 from xmlschema import XMLSchemaParseError, XMLSchemaValidationError
 from xmlschema.validators import XMLSchema11
-from xmlschema.validators.identities import IdentityCounter, KeyrefCounter
+from xmlschema.validators.identities import IdentityCounter, KeyrefCounter, FieldValueSelector
 from xmlschema.testing import XsdValidatorTestCase
 
 
 class TestXsdIdentities(XsdValidatorTestCase):
 
+    cases_dir = pathlib.Path(__file__).parent.joinpath('../test_cases')
+
     def test_key_definition(self):
         schema = self.check_schema("""
             <xs:element name="primary_key" type="xs:string">
@@ -124,22 +127,6 @@ class TestXsdIdentities(XsdValidatorTestCase):
 
         self.assertIn("XPST0003", ctx.exception.message)
 
-    def test_selector_target_namespace(self):
-        schema = self.check_schema("""
-            <xs:element name="primary_key" type="xs:string">
-              <xs:key name="key1">
-                <xs:selector xpath="xs:*"/>
-                <xs:field xpath="."/>
-                <xs:field xpath="@xs:*"/>
-              </xs:key>
-            </xs:element>""")
-
-        self.assertEqual(schema.identities['key1'].selector.target_namespace,
-                         'http://www.w3.org/2001/XMLSchema')
-        self.assertEqual(schema.identities['key1'].fields[0].target_namespace, '')
-        self.assertEqual(schema.identities['key1'].fields[1].target_namespace,
-                         'http://www.w3.org/2001/XMLSchema')
-
     def test_invalid_selector_node(self):
         with self.assertRaises(XMLSchemaParseError) as ctx:
             self.check_schema("""
@@ -322,7 +309,12 @@ class TestXsdIdentities(XsdValidatorTestCase):
                 </xs:complexType>
             </xs:element>""")
 
-        self.assertEqual(schema.identities['key1'].elements, {schema.elements['a']: None})
+        self.assertIn(schema.elements['a'], schema.identities['key1'].elements)
+        self.assertEqual(len(schema.identities['key1'].elements), 1)
+        self.assertTrue(all(
+            isinstance(x, FieldValueSelector)
+            for x in schema.identities['key1'].elements[schema.elements['a']]
+        ))
 
     def test_identity_counter(self):
         schema = self.check_schema("""
@@ -333,7 +325,8 @@ class TestXsdIdentities(XsdValidatorTestCase):
               </xs:key>
             </xs:element>""")
 
-        counter = IdentityCounter(schema.identities['key1'])
+        elem = ElementTree.XML('<primary_key>3</primary_key>')
+        counter = IdentityCounter(schema.identities['key1'], elem)
         self.assertEqual(repr(counter), 'IdentityCounter()')
         self.assertIsNone(counter.increase(('1',)))
         self.assertIsNone(counter.increase(('2',)))
@@ -359,7 +352,8 @@ class TestXsdIdentities(XsdValidatorTestCase):
               </xs:keyref>
             </xs:element>""")
 
-        counter = KeyrefCounter(schema.identities['keyref1'])
+        elem = ElementTree.XML('<primary_key>3</primary_key>')
+        counter = KeyrefCounter(schema.identities['keyref1'], elem)
         self.assertIsNone(counter.increase(('1',)))
         self.assertIsNone(counter.increase(('2',)))
         self.assertIsNone(counter.increase(('1',)))
@@ -370,7 +364,7 @@ class TestXsdIdentities(XsdValidatorTestCase):
         with self.assertRaises(KeyError):
             list(counter.iter_errors(identities={}))
 
-        key_counter = IdentityCounter(schema.identities['key1'])
+        key_counter = IdentityCounter(schema.identities['key1'], elem)
         self.assertIsNone(key_counter.increase(('1',)))
         self.assertIsNone(key_counter.increase('4'))
 
@@ -381,6 +375,18 @@ class TestXsdIdentities(XsdValidatorTestCase):
         self.assertIn("value ('3',) not found", str(errors[1]))
         self.assertIn("(2 times)", str(errors[1]))
 
+    def test_key_multiple_values__issue_418(self):
+        xsd_file = self.casepath('issues/issue_418/issue_418.xsd')
+        schema = self.schema_class(xsd_file)
+
+        xml_file = self.casepath('issues/issue_418/issue_418.xml')
+        self.assertIsNone(schema.validate(xml_file))
+
+        xml_file = self.casepath('issues/issue_418/issue_418-invalid.xml')
+        with self.assertRaises(XMLSchemaValidationError) as ctx:
+            schema.validate(xml_file)
+        self.assertIn("field selects multiple values", str(ctx.exception))
+
 
 class TestXsd11Identities(TestXsdIdentities):
 
@@ -399,8 +405,8 @@ class TestXsd11Identities(TestXsdIdentities):
             </xs:element>""")
 
         key1 = schema.identities['key1']
-        self.assertIsNot(schema.elements['secondary_key'].identities['key1'], key1)
-        self.assertIs(schema.elements['secondary_key'].identities['key1'].ref, key1)
+        self.assertIsNot(schema.elements['secondary_key'].identities[0], key1)
+        self.assertIs(schema.elements['secondary_key'].identities[0].ref, key1)
 
     def test_missing_key_reference_definition(self):
         schema = self.check_schema("""
@@ -451,10 +457,28 @@ class TestXsd11Identities(TestXsdIdentities):
               <xs:keyref ref="keyref1"/>
             </xs:element>""")
 
-        self.assertIsNot(schema.elements['secondary_key'].identities['keyref1'],
-                         schema.elements['primary_key'].identities['keyref1'])
-        self.assertIs(schema.elements['secondary_key'].identities['keyref1'].ref,
-                      schema.elements['primary_key'].identities['keyref1'])
+        self.assertIsNot(schema.elements['secondary_key'].identities[0],
+                         schema.elements['primary_key'].identities[-1])
+        self.assertIs(schema.elements['secondary_key'].identities[0].ref,
+                      schema.elements['primary_key'].identities[-1])
+
+    def test_xsi_type_substitution__issue_446(self):
+        xsd_file = self.casepath('features/substitutions/words.xsd')
+        schema = self.schema_class(xsd_file)
+
+        xml_data = '<words><word>foo</word></words>'
+        self.assertTrue(schema.is_valid(xml_data))
+        xml_data = '<words><word>foo</word><word>foo</word></words>'
+        self.assertFalse(schema.is_valid(xml_data))
+
+        xml_file = self.casepath('features/substitutions/words1.xml')
+        self.assertTrue(schema.is_valid(xml_file))
+        xml_file = self.casepath('features/substitutions/words2.xml')
+        self.assertTrue(schema.is_valid(xml_file))
+        xml_file = self.casepath('features/substitutions/words3.xml')
+        self.assertFalse(schema.is_valid(xml_file))
+        xml_file = self.casepath('features/substitutions/words4.xml')
+        self.assertFalse(schema.is_valid(xml_file))
 
     def test_selector_default_namespace(self):
         schema = self.check_schema("""
@@ -471,9 +495,5 @@ class TestXsd11Identities(TestXsdIdentities):
 
 
 if __name__ == '__main__':
-    import platform
-    header_template = "Test xmlschema's XSD identities with Python {} on {}"
-    header = header_template.format(platform.python_version(), platform.platform())
-    print('{0}\n{1}\n{0}'.format("*" * len(header), header))
-
-    unittest.main()
+    from xmlschema.testing import run_xmlschema_tests
+    run_xmlschema_tests('identities')
