@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -1080,10 +1080,8 @@ class TestXMLSchema11(TestXMLSchema10):
         xml_file = self.casepath('examples/vehicles/vehicles.xml')
 
         schema = self.schema_class(schema_file)
-        with self.assertRaises(XMLSchemaValidationError) as ctx:
-            schema.decode(xml_file, '/vh:vehicles/vh:bikes/vh:bike[2]')
-
-        self.assertIn("maybe you have to provide a different path", ctx.exception.reason)
+        obj = schema.decode(xml_file, '/vh:vehicles/vh:bikes/vh:bike[2]')
+        self.assertEqual(obj, {'@make': 'Yamaha', '@model': 'XS650'})
 
 
 class TestXMLSchemaMeta(unittest.TestCase):
