@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2016-2022, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -33,7 +33,7 @@ from .validators import (
     XMLSchema, XMLSchema10, XMLSchema11, XsdComponent, XsdType, XsdElement, XsdAttribute
 )
 
-__version__ = '4.2.0'
+__version__ = '4.3.0'
 __author__ = "Davide Brunato"
 __contact__ = "brunato@sissa.it"
 __copyright__ = "Copyright 2016-2025, SISSA"
@@ -41,7 +41,7 @@ __license__ = "MIT"
 __status__ = "Production/Stable"
 
 __all__ = [
-    'limits', 'translation', 'settings', 'XMLSchemaException', 'XMLResourceError',
+    'limits', 'translation', 'XMLSchemaException', 'XMLResourceError',
     'XMLSchemaNamespaceError', 'etree_tostring', 'normalize_url',
     'normalize_locations', 'fetch_resource', 'fetch_namespaces',
     'fetch_schema_locations', 'fetch_schema',
