@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2016-2021, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -7,17 +7,17 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
-import os
-
-from ..exceptions import XMLSchemaException, XMLSchemaValueError
-from ..names import XSD_NAMESPACE, WSDL_NAMESPACE, SOAP_NAMESPACE, \
-    SCHEMAS_DIR, XSD_ANY_TYPE, XSD_SCHEMA
-from ..helpers import get_qname, local_name, get_extended_qname, get_prefixed_qname
-from ..namespaces import NamespaceResourcesMap
-from ..resources import fetch_resource
-from ..documents import XmlDocument
-from ..validators import XMLSchema10
+# mypy: ignore-errors
 
+from xmlschema.exceptions import XMLSchemaException, XMLSchemaValueError
+from xmlschema.names import XSD_NAMESPACE, WSDL_NAMESPACE, SOAP_NAMESPACE, \
+    XSD_ANY_TYPE, XSD_SCHEMA
+from xmlschema.utils.qnames import get_qname, local_name, get_extended_qname, \
+    get_prefixed_qname
+from xmlschema.utils.urls import normalize_url
+from xmlschema.locations import SCHEMAS_DIR, get_locations
+from xmlschema.documents import SCHEMA_KWARGS, XmlDocument
+from xmlschema.validators import XMLSchemaBase, XMLSchema10
 
 # WSDL 1.1 global declarations
 WSDL_IMPORT = '{%s}import' % WSDL_NAMESPACE
@@ -72,7 +72,11 @@ class WsdlComponent:
     def __init__(self, elem, wsdl_document):
         self.elem = elem
         self.wsdl_document = wsdl_document
-        self.name = get_qname(wsdl_document.target_namespace, elem.get('name'))
+
+        try:
+            self.name = get_qname(wsdl_document.target_namespace, elem.attrib['name'])
+        except KeyError:
+            self.name = None
 
     def __repr__(self):
         return '%s(name=%r)' % (self.__class__.__name__, self.prefixed_name)
@@ -110,7 +114,7 @@ class WsdlComponent:
 class WsdlMessage(WsdlComponent):
 
     def __init__(self, elem, wsdl_document):
-        super(WsdlMessage, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
         self.parts = {}
         xsd_elements = wsdl_document.schema.maps.elements
         xsd_types = wsdl_document.schema.maps.types
@@ -137,7 +141,7 @@ class WsdlMessage(WsdlComponent):
                     self.parts[part_name] = xsd_elements[element_name]
                 except KeyError:
                     self.parts[part_name] = xsd_types[XSD_ANY_TYPE]
-                    msg = "missing schema element {!r}".format(element_name)
+                    msg = f"missing schema element {element_name!r}"
                     wsdl_document.parse_error(msg)
 
                 continue  # pragma: no cover
@@ -153,14 +157,14 @@ class WsdlMessage(WsdlComponent):
                     self.parts[part_name] = xsd_types[type_name]
                 except KeyError:
                     self.parts[part_name] = xsd_types[XSD_ANY_TYPE]
-                    msg = "missing schema type {!r}".format(type_name)
+                    msg = f"missing schema type {type_name!r}"
                     wsdl_document.parse_error(msg)
 
 
 class WsdlPortType(WsdlComponent):
 
     def __init__(self, elem, wsdl_document):
-        super(WsdlPortType, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
         self.operations = {}
 
         for child in elem.iterfind(WSDL_OPERATION):
@@ -183,7 +187,7 @@ class WsdlOperation(WsdlComponent):
     soap_operation = None
 
     def __init__(self, elem, wsdl_document):
-        super(WsdlOperation, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
         self.faults = {}
 
         input_child = elem.find(WSDL_INPUT)
@@ -243,7 +247,7 @@ class WsdlMessageReference(WsdlComponent):
     message = None
 
     def __init__(self, elem, wsdl_document):
-        super(WsdlMessageReference, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
         message_name = self._parse_reference(elem, 'message')
         try:
             self.message = wsdl_document.maps.messages[message_name]
@@ -287,7 +291,7 @@ class SoapBody(SoapParameter):
     """Class for soap:body bindings."""
 
     def __init__(self, elem, wsdl_document):
-        super(SoapBody, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
         self.parts = elem.get('parts', '').split()
 
 
@@ -300,7 +304,7 @@ class SoapHeader(WsdlMessageReference, SoapParameter):
     part = None
 
     def __init__(self, elem, wsdl_document):
-        super(SoapHeader, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
         if self.message is not None and 'part' in elem.attrib:
             try:
                 self.part = self.message.parts[elem.attrib['part']]
@@ -326,7 +330,7 @@ class WsdlBinding(WsdlComponent):
     """The SOAP binding element if any, `None` otherwise."""
 
     def __init__(self, elem, wsdl_document):
-        super(WsdlBinding, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
         self.operations = {}
 
         if wsdl_document.soap_binding:
@@ -421,7 +425,7 @@ class WsdlPort(WsdlComponent):
     soap_location = None
 
     def __init__(self, elem, wsdl_document):
-        super(WsdlPort, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
 
         binding_name = self._parse_reference(elem, 'binding')
         try:
@@ -440,7 +444,7 @@ class WsdlPort(WsdlComponent):
 class WsdlService(WsdlComponent):
 
     def __init__(self, elem, wsdl_document):
-        super(WsdlService, self).__init__(elem, wsdl_document)
+        super().__init__(elem, wsdl_document)
         self.ports = {}
 
         for port_child in elem.iterfind(WSDL_PORT):
@@ -460,52 +464,89 @@ class Wsdl11Document(XmlDocument):
     """
     Class for WSDL 1.1 documents.
 
-    :param source: a string containing XML data or a file path or an URL or a \
+    :param source: a string containing XML data or a file path or a URL or a \
     file like object or an ElementTree or an Element.
+    :param schema: additional schema for providing XSD types and elements to the \
+    WSDL document. Can be a :class:`xmlschema.XMLSchema` instance or a file-like \
+    object or a file path or a URL of a resource or a string containing the XSD schema.
     :param cls: class to use for building the schema instance (for default \
-    :class:`XMLSchema10` is used).
+    :class:`xmlschema.XMLSchema10` is used).
     :param validation: the XSD validation mode to use for validating the XML document, \
     that can be 'strict' (default), 'lax' or 'skip'.
     :param maps: WSDL definitions shared maps.
     :param namespaces: is an optional mapping from namespace prefix to URI.
     :param locations: resource location hints, that can be a dictionary or a \
     sequence of couples (namespace URI, resource URL).
-    :param base_url: the base URL for base :class:`xmlschema.XMLResource` initialization.
-    :param allow: the security mode for base :class:`xmlschema.XMLResource` initialization.
-    :param defuse: the defuse mode for base :class:`xmlschema.XMLResource` initialization.
-    :param timeout: the timeout for base :class:`xmlschema.XMLResource` initialization.
+    :param kwargs: other optional arguments for initializing :class:`xmlschema.XMLResource` \
+    base class or building :class:`xmlschema.XMLSchema` instances provided as keyword arguments.
     """
     target_namespace = ''
     soap_binding = False
 
-    def __init__(self, source, cls=None, validation='strict', namespaces=None, maps=None,
-                 locations=None, base_url=None, allow='all', defuse='remote', timeout=300):
+    def __init__(self, source, schema=None, cls=None, validation='strict',
+                 namespaces=None, maps=None, locations=None, base_url=None, **kwargs):
 
-        if maps is None:
+        if kwargs.get('lazy'):
+            raise WsdlParseError(f"{self.__class__!r} instance cannot be lazy")
+
+        if maps is not None:
+            self.maps = maps
+            self.schema = maps.wsdl_document.schema
+        else:
             if cls is None:
                 cls = XMLSchema10
-            self.schema = cls(source=os.path.join(SCHEMAS_DIR, 'WSDL/wsdl.xsd'))
+
+            xsd_filepath = f'{SCHEMAS_DIR}WSDL/wsdl.xsd'
+            if isinstance(schema, XMLSchemaBase):
+                self.schema = schema
+            else:
+                self.schema = cls(
+                    source=schema or xsd_filepath,
+                    base_url=base_url,
+                    locations=locations,
+                    **{k: v for k, v in kwargs.items() if k in SCHEMA_KWARGS}
+                )
+
+            self.schema.add_schema(xsd_filepath, WSDL_NAMESPACE, base_url, True)
             self.maps = Wsdl11Maps(self)
-        else:
-            self.schema = maps.wsdl_document.schema
-            self.maps = maps
 
-        if locations:
-            self.locations = NamespaceResourcesMap(locations)
-        else:
-            self.locations = NamespaceResourcesMap()
-
-        super(Wsdl11Document, self).__init__(
+        super().__init__(
             source=source,
             schema=self.schema,
             validation=validation,
             namespaces=namespaces,
             locations=locations,
-            base_url=base_url,
-            allow=allow,
-            defuse=defuse,
-            timeout=timeout,
+            **kwargs,
         )
+        self.target_namespace = self.root.get('targetNamespace', '')
+        self.soap_binding = SOAP_NAMESPACE in self.namespaces.values()
+        self.locations = get_locations(locations, base_url)
+
+        if self.namespace == XSD_NAMESPACE and \
+                self.schema.maps.get_schema(source=self.url) is None:
+            self.schema.__class__(
+                source=self,
+                global_maps=self.schema.maps,
+                locations=self.locations,
+                **{k: v for k, v in kwargs.items() if k in SCHEMA_KWARGS}
+            )
+            return
+
+        if self is self.maps.wsdl_document:
+            self.maps.clear()
+
+        self._parse_imports()
+        self._parse_types()
+        self._parse_messages()
+        self._parse_port_types()
+        self._parse_bindings()
+        self._parse_services()
+
+    def get_arguments(self):
+        """Returns keyword arguments for rebuilding the WSDL document."""
+        kwargs = super().get_arguments()
+        kwargs['locations'] = self.locations
+        return kwargs
 
     @property
     def imports(self):
@@ -532,38 +573,16 @@ class Wsdl11Document(XmlDocument):
         """WSDL 1.1 services."""
         return self.maps.services
 
-    def parse(self, source, lazy=False):
-        if lazy:
-            raise WsdlParseError("{!r} instance cannot be lazy".format(self.__class__))
-
-        super(Wsdl11Document, self).parse(source, lazy)
-        self.target_namespace = self._root.get('targetNamespace', '')
-        self.soap_binding = SOAP_NAMESPACE in self.namespaces.values()
-
-        if self.namespace == XSD_NAMESPACE:
-            self.schema.__class__(self, global_maps=self.schema.maps, locations=self.locations)
-            return
-
-        if self is self.maps.wsdl_document:
-            self.maps.clear()
-
-        self._parse_imports()
-        self._parse_types()
-        self._parse_messages()
-        self._parse_port_types()
-        self._parse_bindings()
-        self._parse_services()
-
     def parse_error(self, message):
-        if self.validation == 'strict':
+        if self._validation == 'strict':
             raise WsdlParseError(message)
-        elif self.validation == 'lax':
+        elif self._validation == 'lax':
             self.errors.append(WsdlParseError(message))
 
     def _parse_types(self):
-        path = '{}/{}'.format(WSDL_TYPES, XSD_SCHEMA)
+        path = f'{WSDL_TYPES}/{XSD_SCHEMA}'
 
-        for child in self._root.iterfind(path):
+        for child in self.root.iterfind(path):
             source = self.subresource(child)
             self.schema.__class__(source, global_maps=self.schema.maps)
 
@@ -571,7 +590,7 @@ class Wsdl11Document(XmlDocument):
         for child in self.iterfind(WSDL_MESSAGE):
             message = WsdlMessage(child, self)
             if message.name in self.maps.messages:
-                self.parse_error("duplicated message {!r}".format(message.prefixed_name))
+                self.parse_error(f"duplicated message {message.prefixed_name!r}")
             else:
                 self.maps.messages[message.name] = message
 
@@ -579,7 +598,7 @@ class Wsdl11Document(XmlDocument):
         for child in self.iterfind(WSDL_PORT_TYPE):
             port_type = WsdlPortType(child, self)
             if port_type.name in self.maps.port_types:
-                self.parse_error("duplicated port type {!r}".format(port_type.prefixed_name))
+                self.parse_error(f"duplicated port type {port_type.prefixed_name!r}")
             else:
                 self.maps.port_types[port_type.name] = port_type
 
@@ -587,7 +606,7 @@ class Wsdl11Document(XmlDocument):
         for child in self.iterfind(WSDL_BINDING):
             binding = WsdlBinding(child, self)
             if binding.name in self.maps.bindings:
-                self.parse_error("duplicated binding {!r}".format(binding.prefixed_name))
+                self.parse_error(f"duplicated binding {binding.prefixed_name!r}")
             else:
                 self.maps.bindings[binding.name] = binding
 
@@ -595,32 +614,30 @@ class Wsdl11Document(XmlDocument):
         for child in self.iterfind(WSDL_SERVICE):
             service = WsdlService(child, self)
             if service.name in self.maps.services:
-                self.parse_error("duplicated service {!r}".format(service.prefixed_name))
+                self.parse_error(f"duplicated service {service.prefixed_name!r}")
             else:
                 self.maps.services[service.name] = service
 
     def _parse_imports(self):
-        namespace_imports = NamespaceResourcesMap(map(
-            lambda x: (x.get('namespace', ''), x.get('location', '')),
-            filter(lambda x: x.tag == WSDL_IMPORT, self.root)
-        ))
+        for child in self.root:
+            if child.tag != WSDL_IMPORT:
+                continue
 
-        for namespace, locations in namespace_imports.items():
-            locations = [url for url in locations if url]
-            try:
+            namespace = child.get('namespace', '').strip()
+            location = child.get('location', '').strip()
+            locations = [location] if location else []
+            if namespace in self.locations:
                 locations.extend(self.locations[namespace])
-            except KeyError:
-                pass
 
             import_error = None
             for url in locations:
                 try:
                     self.import_namespace(namespace, url, self.base_url)
-                except (OSError, IOError) as err:
+                except OSError as err:
                     if import_error is None:
                         import_error = err
                 except SyntaxError as err:
-                    msg = "cannot import namespace %r: %s." % (namespace, err)
+                    msg = f"can't import namespace {namespace!r}: {err}."
                     self.parse_error(msg)
                 except XMLSchemaValueError as err:
                     self.parse_error(err)
@@ -641,12 +658,12 @@ class Wsdl11Document(XmlDocument):
         elif namespace in self.maps.imports:
             return self.maps.imports[namespace]
 
-        url = fetch_resource(location, base_url or self.base_url)
+        url = normalize_url(location, base_url or self.base_url)
         wsdl_document = self.__class__(
             source=url,
             maps=self.maps,
-            namespaces=self._namespaces,
-            validation=self.validation,
+            namespaces=self._init_namespaces,
+            validation=self._validation,
             base_url=self.base_url,
             allow=self.allow,
             defuse=self.defuse,
